using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.IO;

using Pencil.Gaming;
using Pencil.Gaming.Audio;
using Pencil.Gaming.Graphics;
using Pencil.Gaming.MathUtils;
using Pencil.Gaming.Scripting;

namespace LudumDare26 {
	public class BillBoardedSprite : Texture2D {
		private readonly Program program;

		protected virtual Enemy ProvideEnemy() {
			return null;
		}

		public float X { get; protected set; }
		public float Z { get; protected set; }
		public float Y { get; protected set; }
		public float GameWidth { get; private set; }
		public float GameHeight { get; private set; }
		public bool IsExposed { get; private set; }

		public BillBoardedSprite(Program prgrm, float x, float y, float z, float wGame, float hGame, int width, int height) : base(width, height) {
			program = prgrm;
			X = x;
			Z = z;
			Y = y;
			GameWidth = wGame;
			GameHeight = hGame;
		}

		public void Draw() {
			float rot = program.Rotation;
			float rCos = (float)Math.Cos(rot);
			float rSin = (float)Math.Sin(rot);

			float xTrans = X + program.PlayerX - GameWidth / 2f;
			float zTrans = Z - program.PlayerZ;

			float xSprite = xTrans * rCos - zTrans * rSin;
			float zSprite = zTrans * rCos + xTrans * rSin;

			xSprite += GameWidth / 2f;

			if (zSprite >= 0f) {
				if (xSprite * xSprite + zSprite * zSprite < Program.ClipDistance * Program.ClipDistance) {
					//float yTopScreen = Math3D.GetScreenY2(zSprite);
					float yBottomScreen = Math3D.GetScreenY1(zSprite);
					float heightPixelsOver2 = Math.Abs(yBottomScreen - Program.PixelHeight / 2f);
					float yTopScreen = yBottomScreen - 2 * GameHeight * heightPixelsOver2;

					float xScreenLeft = Math3D.GetScreenX(xSprite + 0, yBottomScreen);
					float xScreenRight = Math3D.GetScreenX(xSprite - GameWidth, yBottomScreen);

					if (xScreenLeft > xScreenRight) {
						float temp = xScreenLeft;
						xScreenLeft = xScreenRight;
						xScreenRight = temp;
					}

					float zBuff = 1f;
					if (yBottomScreen < Program.PixelHeight) {
						zBuff = ((Program.PixelHeight / 2f) / heightPixelsOver2);
					}
			
					IsExposed = false;
					for (int x = (int)xScreenLeft; x < (int)xScreenRight; ++x) {
						if (x >= 0 && x < Program.PixelWidth) {
							for (int y = (int)yTopScreen; y < (int)yBottomScreen; ++y) {
								if (y >= 0 && y < Program.PixelHeight) {
									if (Math.Abs(zBuff) < program.GetZBufferPixel(x, y)) {
										float tCoordX = Math.Abs((x - xScreenLeft) / (float)(xScreenRight - xScreenLeft));
										float tCoordY = Math.Abs((y - yTopScreen) / (float)(yBottomScreen - yTopScreen));

										int tCoordXInt = (int)(tCoordX * Width);
										int tCoordYInt = (int)(tCoordY * Height);

										uint color = GetPixel(tCoordXInt, tCoordYInt);
										program.SetPixel(x, y, color);
										program.SetZBufferPixel(x, y, Math.Abs(zBuff));
										program.SetForegroundBufferPixel(x, y, ProvideEnemy());
									}
								}
							}
						}
					}
				}
			}
		}
	}
}

