using System;

namespace LudumDare26 {
	public class Enemy : BillBoardedSprite {
		private float counter = 0f;
		private readonly Program program;
		private bool flickering = false;
		private float flickerCounter = 0f;

		protected override Enemy ProvideEnemy() {
			return this;
		}

		public Enemy(Program prog, float x, float z) : base(prog, x, 0f, z, .5f, .9f, 1, 1) {
			program = prog;
			Clear(0xFFFF0000);
		}

		public bool CanShoot() {
			float gridTileX = X;
			float gridTileZ = Z;

			float playerGridTileX = -program.PlayerX;
			float playerGridTileZ = program.PlayerZ;

			float deltaX = gridTileX - playerGridTileX;
			float deltaZ = gridTileZ - playerGridTileZ;

			const float iterations = 10;

			float xStep = deltaX / iterations;
			float zStep = deltaZ / iterations;

			float xLoop = gridTileX;
			float zLoop = gridTileZ;
			for (int i = 0; i < iterations; ++i) {
				float xStepI = xStep * i;
				float zStepI = zStep * i;

				float xLoopAdj = xLoop - xStepI;
				float zLoopAdj = zLoop - zStepI;

				int xLoopTile = (int)Math.Ceiling(xLoopAdj);
				int zLoopTile = (int)Math.Floor(zLoopAdj);

				foreach (Block b in program.Blocks) {
					if (b.X == -xLoopTile && b.Z == zLoopTile) {
						return false;
					}
				}
			}

			return true;
		}

		public void Shoot() {
			program.HurtPlayer();
			program.PlayerHealth -= .05f;
			counter = 0f;

			flickering = true;
			flickerCounter = 1f;
		}

		public void Update(float time) {
			if (flickering) {
				flickerCounter -= time;
				Clear(UintFromColor(
					255, 
					(byte)(flickerCounter * 10f * 255 % 255), 
					(byte)(flickerCounter * 10f * 255 % 255), 
					(byte)(flickerCounter * 10f * 255 % 255))
				);
				if (flickerCounter < 0f) {
					flickering = false;
				}
			} else {
				Clear(0xFFFF0000);

			}

			if ((X + program.PlayerX) * (X + program.PlayerX) + (Z - program.PlayerZ) * (Z - program.PlayerZ) < Program.ClipDistance * Program.ClipDistance) {
				counter += time;
			
				if (counter > 2) {
					if (CanShoot()) {
						Shoot();
					}
					counter = 0f;
				}
			}
		}
	}
}
