using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;

using Pencil.Gaming;
using Pencil.Gaming.Audio;
using Pencil.Gaming.Graphics;
using Pencil.Gaming.MathUtils;
using Pencil.Gaming.Scripting;

namespace LudumDare26 {
	public class Wall : Texture2D {
		private readonly Program program;
		private int x, z;
		private bool alongX;

		public Wall(Program p, int x, int z, bool alongX) : base(32, 32) {
			program = p;

			this.x = x;
			this.z = z;
			this.alongX = alongX;

			Clear(0xFFFFFFFF);
		}

		/*private void ConnectPoints3D(Vector2i p0, Vector2i p1, Vector2i p3, float xWall, float zWall) {
			float dy_dx = (p1.Y - p0.Y) / (float)(p1.X - p0.X);

			int minx = Math.Min(p0.X, p1.X);
			int maxx = Math.Max(p0.X, p1.X);
			int minx1 = minx;
			int maxx1 = maxx;
			if (xWall > 0) {
				int temp = maxx1;
				maxx1 = minx1;
				minx1 = temp;
			}
			for (int x = minx; x <= maxx; ++x) {
				for (int y = Math.Max((int)(p3.Y - dy_dx * (x - maxx1)), 0); y < Math.Min((int)(p0.Y + dy_dx * (x - minx1)), Program.PixelHeight); ++y) {
					if (x >= 0 && x < Program.PixelWidth && y >= 0 && y < Program.PixelHeight) {
						float zBuff = (Program.PixelHeight / 2f) / (p0.Y + dy_dx * (x - minx1) - Program.PixelHeight / 2f);
						if (p0.Y + dy_dx * (x - minx1) > Program.PixelHeight) {
							zBuff = 1f;
						}
						if (zBuff < program.GetZBufferPixel(x, y)) {
							program.SetPixel(x, y, 0xFFFFFFFF);
							program.SetZBufferPixel(x, y, zBuff);
						}
					}
				}
			}
		}

		private void DrawWallAlongX(float xWall, float zWall) {
			float rotCos = (float)Math.Cos(program.Rotation);
			float rotSin = (float)Math.Sin(program.Rotation);
			float zWallBackup = zWall;

			//zWall = rotCos * zWall + rotSin * xWall;
			//xWall = rotCos * xWall + rotSin * zWallBackup;

			zWall -= program.PlayerZ;
			xWall -= program.PlayerX;

			if (zWall > .1f) {
				float y1 = GetScreenY1(zWall);
				float y3 = GetScreenY1(zWall + rotSin);
				float y4 = GetScreenY2(zWall + rotSin);

				Vector2i p0 = new Vector2i((int)GetScreenX(xWall, y1), (int)y1);
				Vector2i p1 = new Vector2i((int)GetScreenX(xWall + 0f, y3), (int)y3);
				Vector2i p3 = new Vector2i((int)GetScreenX(xWall + 0f, y3), (int)y4);

				ConnectPoints3D(p0, p1, p3, xWall, zWall);
			}
		}

		private void DrawWallAlongY(float xWall, float zWall) {
			float rotCos = (float)Math.Cos(program.Rotation);
			float rotSin = (float)Math.Sin(program.Rotation);
			float zWallBackup = zWall;

			//zWall = rotCos * zWall + rotSin * xWall;
			//xWall = rotCos * xWall + rotSin * zWallBackup;

			zWall -= program.PlayerZ;
			xWall -= program.PlayerX;

			if (zWall > .1f) {
				float y1 = GetScreenY1(zWall);
				float y3 = GetScreenY1(zWall);
				float y4 = GetScreenY2(zWall);

				Vector2i p0 = new Vector2i((int)GetScreenX(xWall, y1), (int)y1);
				Vector2i p1 = new Vector2i((int)GetScreenX(xWall, y3), (int)y3);
				Vector2i p3 = new Vector2i((int)GetScreenX(xWall, y3), (int)y4);

				ConnectPoints3D(p0, p1, p3, xWall, zWall);
			}
		}*/


		private void DrawWallRotated(float xWall, float zWall, float rot) {
			//rot = 4.48f;
			//rot = 200;

			float rCos = (float)Math.Cos(rot);
			float rSin = (float)Math.Sin(rot);

			float rCosWorld = (float)Math.Cos(program.Rotation);
			float rSinWorld = (float)Math.Sin(program.Rotation);

			xWall += program.PlayerX; // Don't ask me why it's reversed, it just is
			zWall -= program.PlayerZ;

			float zWallBackup = zWall;
			float xWallBackup = xWall;
			xWall = rCosWorld * xWall - rSinWorld * zWall;
			zWall = rCosWorld * zWallBackup + rSinWorld * xWallBackup;

			if (xWall * xWall + zWall * zWall < Program.ClipDistance * Program.ClipDistance) {
				if (zWall >= 0f) {
					Vector2 origin = new Vector2(xWall, zWall);
					Vector2 otherPart = origin + new Vector2(-rCos, rSin); // See the note given above

					float yOriginTop = Math3D.GetScreenY2(origin.Y);
					float yOriginBottom = Math3D.GetScreenY1(origin.Y);
					float yOtherTop = Math3D.GetScreenY2(otherPart.Y);

					Vector2i pOriginTop = new Vector2i((int)Math3D.GetScreenX(origin.X, yOriginTop), (int)yOriginTop);
					Vector2i pOriginBottom = new Vector2i(pOriginTop.X, (int)yOriginBottom);
					Vector2i pOtherTop = new Vector2i((int)Math3D.GetScreenX(otherPart.X, yOtherTop), (int)yOtherTop);

					ConnectPoints3D(pOriginTop, pOriginBottom, pOtherTop);
				}
			}
		}

		private void ConnectPoints3D(Vector2i pOriginTop, Vector2i pOriginBottom, Vector2i pOtherTop) {
			float dy_dx = (pOtherTop.Y - pOriginTop.Y) / (float)(pOtherTop.X - pOriginTop.X);
			//float dy_dx = .1f;
			//dy_dx = -dy_dx;

			int xMin = Math.Min(pOriginTop.X, pOtherTop.X);
			int xMax = Math.Max(pOriginTop.X, pOtherTop.X);

			int yMin;
			int yMax;

			if (dy_dx >= 0) {
				yMin = pOriginTop.Y;
				yMax = pOriginBottom.Y;
			} else {
				yMin = pOtherTop.Y;
				yMax = (int)(pOriginBottom.Y - dy_dx * (pOtherTop.X - pOriginTop.X));
			}

			if (pOriginTop.Y > pOtherTop.Y) {
				if (dy_dx < 0) {
					yMin = pOriginTop.Y;
					yMax = pOriginBottom.Y;
				} else {
					yMin = pOtherTop.Y;
					yMax = (int)(pOriginBottom.Y - dy_dx * (pOtherTop.X - pOriginTop.X));
				}
			}

			for (int x = xMin; x <= xMax; ++x) {
				// Draw strips
				if (x >= 0 && x < Program.PixelWidth) {
					int xStep = x - xMin;
					for (int y = (int)(yMin + (dy_dx * xStep)); y <= yMax - (dy_dx * xStep); ++y) {
						if (y >= 0 && y < Program.PixelHeight) {
							float zBuff = (Program.PixelHeight / 2f) / (yMin + dy_dx * xStep - Program.PixelHeight / 2f);
							if (yMax - dy_dx * xStep > Program.PixelHeight - 1) {
								zBuff = 1f;
							}
							if (Math.Abs(zBuff) < program.GetZBufferPixel(x, y)) {
								program.SetPixel(x, y, 0xFFFFFFFF);
								program.SetZBufferPixel(x, y, Math.Abs(zBuff));
								program.SetForegroundBufferPixel(x, y, null);
							}
						}
					}
				}
			}
		}

		public void Draw() {
			if (alongX) {
				//DrawWallAlongX(x, z);
				DrawWallRotated(-(x), z, -program.Rotation);
				//DrawWallAlongY(x 1, z);
			} else {
				//DrawWallAlongY(x, z);
				DrawWallRotated(-(x), z, MathHelper.TauOver4 - program.Rotation);
			}
		}
	}
}

